/* IBM Specific parts of DCALC.C - Pop-up HP Calculator.
*
*  DCALC is an RPN (ie HP) calculator that can terminate and stay resident.
*  It uses windows (in text mode only) so that it does not interfere with
*  existing applications. It can read a rectangular region of figures from the
*  screen and add them up and, later, send its results to the keyboard.
*
*  DCALC requests that the scheduler give it control every timer
*  tick, so that if it can only stuff a portion of its
*  output string, it can stuff more of it on the next timer
*  tick, until the whole string has been stuffed.
*
*  If DCALC receives a hot key when another hot key is being
*  processed, the second hot key is ignored.
*
*  DCALC retrieves the hot key pressed from the IV_EVENT structure
*  passed to it by the intervention code scheduler.
*
*  The command line format is as follows:
*
*       DCALC [-R -I]
*
*       -R: remove from memory
*       -I: install in memory (TSR - default is to terminate normally)
*
*  DCALC is a self-removing interrupt service routine.  If DCALC is
*  invoked with the "-R" argument, it removes a running copy of DCALC
*  (if present).
*
*  Default hot keys -
*     ALT-1 Pop up
*     ALT-2 Type X register into current application
*     ALT-3 capture column of numbers from (text) screen, perform arithmetic
*
**/

#include <stdio.h>
#include <ctype.h>
#include <dos.h>
#include <setjmp.h>
#include <signal.h>

#ifdef TSR
#include <binterv.h>
#endif /* TSR */

#include <bintrupt.h>
#include <bkeybrd.h>
#include <bkeys.h>
#include <bwindow.h>
#include <bvideo.h>

#include "dcalc.h"

#if PCSPECIFIC

/* This code depends on TURBO C 1.5 */
#ifdef __TURBOC__

#ifndef __SMALL__
  you lose ... compile this and DCALC.C in the small model please
#endif

#define NUM_COMMANDS	5
#define ESC     	'\033'
#define CR      	'\r'
#define PLUS    	'+'
#define MINUS   	'-'
#define DIVIDE  	'/'
#define MULT    	'*'
#define NO_BUTTON	0
#define LEFT_BUTTON	1
#define RIGHT_BUTTON	2
#define STKSIZE  	4000
#define OK       	0
#define FALL_OUT 	101
#define NOT_FOUND         1
#define ERROR_DISABLING   2
#define ERROR_REMOVING    3
#define ALREADY_INSTALLED 4
#define SCREEN_MAX_X	79
#define SCREEN_MAX_Y	24
#define k(c) kbscanof(c)

	/* Definitive signature for this version of DCALC.  */
#define DCALC_SIGN "DCALC 26/06/91"

extern int 		base,
			fk_set;
extern struct funckey 	func[NUMFK][8];
extern long 		x, y, z, t, l;
extern double 		xf, yf, zf, tf, lf;

/* TURBOC gives you 64K of heap unless you specify an amount ... */
unsigned int		_heaplen = 3000;

static jmp_buf		save;		/* Used in dcalc() to recover disaster*/
static int 		cur_y, cur_x,	/* Set by gotoxy, used by print_string*/
			cursor = 0,	/* Is cursor on? Used by cursor_on/off*/
			region = 0,	/* Is region on? Used by region_on/off*/
			x1old, y1old,
			x2old, y2old;	/* Cursor & region coordinates	      */
static char 		*pstuff = NULL; /* Used by tickhandler unless NULL */
static char 		outbuf[45];
static char 		*outptr = outbuf;
static char		reg_mem_msg[] =
				"Not enough memory for register display",
			hlp_mem_msg[] =
				"Not enough memory for help display";

int          		is_resident = FALSE;
#ifdef TSR
static int   		awake  = TRUE;

	/* Allocation of stack space for the intervention   */
	/* scheduler and user function.                     */

char			schedstk[STKSIZE];
#endif /* TSR */
int     		cursor_was_off,
			old_row, old_col, old_high, old_low;

/* Forward declarations */
COMMAND			get_a_char();

#ifdef TSR
void 			tickhandler();
int 			install_iv();
int 			remove_iv();
#endif /* TSR */

void			os_init();

/* WINDOW STORAGE - PREALLOC'D IN 'MAIN' */

WHERE		xlocation,
		reglocation,
		msglocation,
		fklocation,
		helplocation;

BWINDOW         *xwin, *helpwin, *msgwin, *regwin, *fkwin;

BORDER 		xbord = {
			BBRD_DDDD | BBRD_TLT,
			MAGENTA,
			' ', ' ',
			"DCALC - Use \'H\' for Help",
			NULL,
			REVERSE
		},
		regbord = {
			BBRD_DDDD | BBRD_TLT,
			MAGENTA,
			' ', ' ',
			"DCALC: \'@\' turns registers off",
			NULL,
			REVERSE
		},
		msgbord = {
			BBRD_DDDD | BBRD_TLT,
			MAGENTA,
			' ', ' ',
			"DCALC:",
			NULL,
			REVERSE
		},
		fkbord = {
			BBRD_NO_BORDER,
			MAGENTA,
			' ', ' ',
			NULL,
			NULL,
			REVERSE
		},
		helpbord = {
			BBRD_DDDD | BBRD_TLT,
			MAGENTA,
			' ', ' ',
			"DCALC: \'H\' turns HELP window off",
			NULL,
			REVERSE
		};

int usage()
{
	puts("Usage: DCALC [-R -I]");
	puts("-I   :installs DCALC in memory");
	puts("-R   :removes DCALC from memory");
	return(1);
}

/***********************************/
/* TSR & Pop-up specific stuff ... */
/***********************************/

#ifdef TSR

struct options {
	char  signature[sizeof(DEF_SIG)];
	IV_KEY keytab[NUM_COMMANDS];
	int   chksum;
};

struct options Opt = {
	DEF_SIG,
	{ {KB_C_A_1,  KB_S_A_1,  IV_KY_SERVICE}, /* ALT-1 Pop-up the calculator */
	  {KB_C_A_2,  KB_S_A_2,  IV_KY_SERVICE}, /* ALT-2 Send X to underlying program */
	  {KB_C_A_3,  KB_S_A_3,  IV_KY_SERVICE}, /* ALT-3 Window capture */

		/* Sleep and wake hot keys */
	  {KB_C_A_MINUS,  KB_S_A_MINUS,  IV_KY_SLEEP  },
	  {KB_C_A_EQUALS, KB_S_A_EQUALS, IV_KY_WAKE   },
	},
	0 /* Checksum */
};

	/* This is the data structure to pass to the        */
	/* scheduler so that it will give control every     */
	/* clock tick.                                      */
IV_TIME timetab   = {1, IV_TM_INTERVAL};

	/* This is the data structure to pass to the        */
	/* scheduler so that it will give control every     */
	/* time a hot key is pressed.                       */

/**
*
* Name          INSTALL_IV -- Install interrupt vectors for DCALC, then
*                             terminate and stay resident.
*
* Synopsis      ret = install_iv ();
*
*               int ret         Return code from IVINSTAL if an
*                               error condition was encountered.
*
* Description   This function installs DCALC if another copy of
*               DCALC is not already installed.
*
* Returns       ret     ALREADY_INSTALLED (4)
*                           A copy of DCALC is already installed.
*                       FALL_OUT (101)
*                           ISRESEXT() failed.
*                       IV_INSTALLED (5)
*                           A copy of DCALC is already installed.
*
**/

int install_iv ()
{
	int        ercode;
	IV_VECTORS vecs;

	/* Check to see if DCALC already installed.         */
	ivvecs (IV_RETVEC, &vecs);
	if (ivsense (&vecs, DCALC_SIGN) != FARNIL) {
		puts ("DCALC already installed.");
		return (ALREADY_INSTALLED);
	}


	/* Install the interrupt service routine--i.e. tell */
	/* the scheduler about our tick and keypress        */
	/* handler routine.                                 */
	if (0 !=
		(ercode =
		 ivinstal (tickhandler, DCALC_SIGN, schedstk, STKSIZE,
			 Opt.keytab,
			 NUM_COMMANDS,
			 &timetab,
			 sizeof(timetab) / sizeof(IV_TIME),
			 IV_DOS_NEED |
			 IV_DKEY_NEED | IV_FLOAT_NEED))) {
	/* Error!                                           */
				printf ("IVINSTAL error %d.\n", ercode);
				return (ercode);
			}

	puts(Opt.signature);
	puts(" ... installed\n");
	puts("ALT-1 to pop up");
	puts("ALT-2 to export the X register to the current application");
	puts("ALT-3 to use a mouse (or keys) to gather numbers\n");
	puts("After ALT-3, use right mouse button or ESC to cancel the gathering.");
	puts("Otherwise, use the left mouse button, ENTER or one of +-/* to mark 2 corners");
	puts("of the region to be gathered and added (or subtracted etc).");
	puts("\nBug City - when using with some programs (eg COMMAND.EXE/COM), you need to");
	puts("press any key (eg space) after ALT-1,2 or 3.");

	/* Terminate and stay resident.                     */
	is_resident = TRUE;
	isresext (OK);

	/* Should never get here.                           */
	return (FALL_OUT);
}

/**
*
* Name          REMOVE_IV -- Remove a previously installed copy of DCALC.
*
* Synopsis      ret = remove_iv ();
*
*               int ret         Return code.
*                                 NO_ERROR  (0)-
*                                     No error encountered.
*                                 NOT_FOUND (1)-
*                                     DCALC not found.
*                                 ERROR_DISABLING (2)-
*                                     DCALC could not be disabled.
*                                     This error should *never* be
*                                     seen.
*                                 ERROR_REMOVING (3)--
*                                     DCALC could not be removed (most
*                                     likely overwritten MALLOC
*                                     pointers).
*
* Description   This function removes a currently-active copy of DCALC
*               from memory, restoring interrupt vectors and freeing
*               memory in the process.
*
* Returns       ret (nonzero if error--see above).
*
**/

int remove_iv ()
{
	IV_VECTORS   vecs;
	IV_CTRL far *pivctrl;

	/* Check to see if DCALC installed.                 */
	ivvecs (IV_RETVEC, &vecs);
	if ((pivctrl = ivsense (&vecs, DCALC_SIGN)) == FARNIL) {
		puts ("DCALC was not in memory!");
		return (NOT_FOUND);
	}

	if (ivdisabl (pivctrl)) {
		puts ("Error disabling DCALC.");
		return (ERROR_DISABLING);
	}

	if (isremove (pivctrl->psp)) {
		puts ("Error removing DCALC.");
		return (ERROR_REMOVING);
	}

	puts("DCALC removed from memory.");
	return (0);
}

/**
*
* Name          TICKHANDLER -- Handle clock tick and hot key
*                              events.
*
* Synopsis      (To be called only by intervention code scheduler)
*
*               tickhandler (pevent);
*
*               IV_EVENT *pevent        Pointer to intervention event
*                                       structure.  The structure
*                                       contains the hot key which
*                                       was detected (if any), as
*                                       well as other data.
*
* Description   This function accepts control from the scheduler
*               every timer tick, and also every time one of the
*               defined hot keys is pressed.  The two cases are
*               dissimilar:
*                   1.  Hot key: Ignore it if we are already
*                       processing another hot key.  If not, DCALC
*                       stuffs as much of the hot key expansion
*                       as it can into the keyboard buffer, then
*                       leaves a pointer to the remainder for the
*                       timer tick invocation to clean up.
*                   2.  Timer tick: DCALC tries to stuff any
*                       remaining hot key expansions into the
*                       keyboard buffer.  If it cannot stuff the
*                       complete string, it schedules the rest for
*                       the next timer tick.
*
* Returns       None.
*
**/

void tickhandler (pevent)
IV_EVENT *pevent;
{
	void dcalc();
	void type_x();
	void capture();

	/* Do a key action, if one exists.                  */
	switch (pevent->key.action) {
	/* If there is no key action, ignore the hot key    */
	/* part of the handler.                             */
		case IV_KY_NONE:
			break;

		case IV_KY_SERVICE:
		/* If we are already stuffing a string, discard this*/
		/* hot key.  If we are asleep, ignore it.           */
			if ((pstuff == NULL) && awake) {
				if (pevent->key.keycode == Opt.keytab[0].keycode) {
					os_init();
					dcalc();	/* Pop-up the calculator */
				} else if (pevent->key.keycode == Opt.keytab[1].keycode) {
					type_x(outbuf);	/* Write X to the keyboard */
					pstuff = outbuf;
				} else if (pevent->key.keycode == Opt.keytab[2].keycode) {
					os_init();
					if (setjmp(save) == 0) {
						set_signals();
						capture();	/* Use mouse to gather numbers */
					}
				}
			}
			break;

		/* If we get a message to go to sleep, do it.       */
		case IV_KY_SLEEP:
			awake = FALSE;
			utsound (800, 1);
			break;

		/* IF we get a message to wake up, do it.           */
		case IV_KY_WAKE:
			awake = TRUE;
			utsound (1600, 1);
			break;
	}

	if (pstuff != NULL) {
		/* Use KBPLACE to stuff them in, character by       */
		/* character, until full.                           */
		while ((*pstuff != ASCIIZ) &&
			 (kbplace (KB_TAIL, (int) *pstuff, kbscanof (*pstuff)) == 0))
			pstuff++;

		/* If we have come to the end of the string, we     */
		/* must set pstuff to NULL so that we know we can    */
		/* accept another hot key now.                      */
		if (*pstuff == ASCIIZ) {
			*outbuf = ASCIIZ;
			outptr = outbuf;
			pstuff = NULL;
		}
	}
}

/********************/
/* End of TSR stuff */
/********************/

/* Type the value currently in the calculator's X register to the keyboard -
 * only called when DCALC is TSR - the underlying application gets the number.
 */
void type_x(s)
	char 	*s;
{
	char 	*d, *c;

	fmt_base(s, x, xf);

	/* remove leading blanks */
	while (*s && isspace(*s)) {
		c = s;
		d = s + 1;
		while (*c)
			*c++ = *d++;
	}
}

/*****************************************************/
/* This section for capturing a region on the screen */
/*****************************************************/

void invert_char(x, y)
	int	x, y;
{
	char	b[2];
	int	fore, back;

	virdrect(y, x, y, x, b, CHAR_ATTR);
	fore = ((b[1] >> 4) & 7) | (b[1] & 8);
	back = (b[1] & 7) | ((b[1] >> 4) & 8);
	viatrect(y, x, y, x, fore, back);
}

void cursor_on(x, y)
	int	x, y;
{
	if (cursor && (x == x1old) && (y == y1old))
		return;

	if (cursor)
		invert_char(x1old, y1old);
	invert_char(x, y);
	x2old = x1old = x;
	y2old = y1old = y;
	cursor = 1;
}

void cursor_off()
{
	if (cursor) {
		invert_char(x1old, y1old);
		cursor = 0;
	}
}

void region_on(x1, y1, x2, y2)
	int	x1, y1, x2, y2;
{
	int	x, y, t,
		minx1, maxx1,
		miny1, maxy1,
		minx2, maxx2,
		miny2, maxy2;

	if (x1 > x2) {
		t = x1;
		x1 = x2;
		x2 = t;
	}
	if (y1 > y2) {
		t = y1;
		y1 = y2;
		y2 = t;
	}

	if (region && (x1 == x1old) && (y1 == y1old) &&
		      (x2 == x2old) && (y2 == y2old))
		return;

	/* Undraw old except where it is inside new region */
	if (region)
		for (y = y1old; y <= y2old; y++)
			for (x = x1old; x <= x2old; x++)
				if ((x >= x1) && (x <= x2) && (y >= y1) && (y <= y2))
					;
				else
					invert_char(x, y);

	/* Draw new except where it is inside old region */
	for (y = y1; y <= y2; y++)
		for (x = x1; x <= x2; x++)
			if (region && (x >= x1old) && (x <= x2old) && (y >= y1old) && (y <= y2old))
				;
			else
				invert_char(x, y);

	x1old = x1;
	y1old = y1;
	x2old = x2;
	y2old = y2;
	region = 1;
}

void region_off()
{
	int	x, y;

	if (region)
		for (y = y1old; y <= y2old; y++)
			for (x = x1old; x <= x2old; x++)
				invert_char(x, y);

	region = 0;
}

 /* Get mouse position & button status */
int get_mouse(i, j)
	int 	*i, *j;
{
	union REGS inregs, outregs;

	inregs.x.ax = 3;
	int86(51, &inregs, &outregs); /* Get mouse position & button status */
	*i = outregs.x.cx >> 3;
	*j = outregs.x.dx >> 3;
	return(outregs.x.bx);
}

/* Set mouse position */
set_mouse(x, y)
	int	x, y;
{
	union REGS inregs, outregs;

	inregs.x.ax = 4;
	inregs.x.cx = x << 3;
	inregs.x.dx = y << 3;
	int86(51, &inregs, &outregs); /* set cursor position */
}

/* Get a key and act on it ... only called from get_region */
int echo_arrow(i, j)
	int 	*i, *j;
{
	COMMAND	c, d;

	c = get_a_char(&d);
	if (c == 0) {
		switch(d) {
			case UPKEY  :  (*j)--;             break;
			case DOWNKEY:  (*j)++;             break;
			case LEFTKEY:  (*i)--;             break;
			case RIGHTKEY: (*i)++;             break;
			case HOMEKEY:  (*i) = 0;  *j = 0;  break;
			case PREVKEY:
				(*i) = SCREEN_MAX_X;
				*j = 0;
				break;
			case NEXTKEY:
				(*i) = SCREEN_MAX_X;
				*j = SCREEN_MAX_Y;
				break;
			case ENDKEY:
				(*i) = 0;
				*j = SCREEN_MAX_Y;
				break;
			default:			   break;
		}
		if (*j > SCREEN_MAX_Y)	*j = SCREEN_MAX_Y;
		if (*j < 0)		*j = 0;
		if (*i > SCREEN_MAX_X)	*i = SCREEN_MAX_X;
		if (*i < 0)		*i = 0;
	}
	return(c);
}

int get_region(x1, y1, x2, y2)
	int *x1, *y1, *x2, *y2;
{
	char		c;
	union REGS	inregs, outregs;
	int 		mouse_present = 0;
	int 		button, d;

	button = 0;

	/* Detect mouse - if not there, use the direction keys */
	inregs.x.ax = 0;
	int86(51, &inregs, &outregs);
	if (outregs.x.ax != 0) {
		/* num_buttons = outregs.x.bx; .... in case I ever need it */
		mouse_present = 1;

		inregs.x.ax = 1;
		/* int86(51, &inregs, &outregs); /* Turn on sprite */

		set_mouse(*x1, *y1);
		/* wait for buttons up (in case he's left his finger on the tit) */
		do
			button = get_mouse(&d, &d);
		while (button != 0);
	}

	do {
		/* Get first point */
		cursor_on(*x1, *y1);
		button = 0;
		while (button == 0) {
			if (mouse_present)
				button = get_mouse(x1, y1);

			if (kbhit()) {
				button = echo_arrow(x1, y1);
				set_mouse(*x1, *y1);
			}

			cursor_on(*x1, *y1);
		}

		cursor_off();

		/* If using mouse, wait for button release */
		while (get_mouse(&d, &d))
			;

		/* Get second point */
		if ((button == 1) || (button == CR)) {
			*x2 = *x1;
			*y2 = *y1;
			region_on(*x1, *y1, *x2, *y2);

			button = 0;
			while (button == 0) {
				if (mouse_present)
					button = get_mouse(x2, y2);

				if (kbhit()) {
					button = echo_arrow(x2, y2);
					set_mouse(*x2, *y2);
				}
				region_on(*x1, *y1, *x2, *y2);
			}
		}

		region_off();

		/* If using mouse, wait for button release */
		while (get_mouse(&d, &d))
			;

	} while ((button != 1) &&
		 (button != 2) &&
		 (button != CR) &&
		 (button != PLUS) &&
		 (button != MINUS) &&
		 (button != DIVIDE) &&
		 (button != MULT) &&
		 (button != ESC));

	return(button);
}

/* Scan from x1 to x2 on line y1. If x2 reached then go to next line. */
int get_value(x1save, x1, y1, x2, y2, value)
	int	x1save, *x1, *y1, x2, y2;
	double	*value;
{
	char	buf[81],
		*s;
	extern	char far *viptr();
	int 	dot_legal =	1,
		exp_legal =	1,
		sign_legal =	1,
		comma_legal =	1,
		number_found =	0;

	s = buf;

	/* skip until we hit a number */
	while ((*y1 <= y2) &&
	       (!isdigit(*s = *viptr(*y1, *x1)) ) &&
	       (*s != '-') &&
	       (*s != '.')) {
		(*x1)++;
		if (*x1 > x2) {
			*x1 = x1save;
			(*y1)++;
		}
	}

	while ((*y1 <= y2) &&
	       (*x1 <= x2) && /* end of line terminates the number */
	       (
		(isdigit(*s = *viptr(*y1, *x1)))      ||
		((*s == ',') && comma_legal)          ||
		((*s == '.') && dot_legal)            ||
		((toupper(*s) == 'E') && exp_legal)   ||
		((*s == '-') && sign_legal)
	       )) {
		if (*s == '.') {
			dot_legal = FALSE;
			sign_legal = FALSE;
			comma_legal = FALSE;
		}
		if (isdigit(*s)) {
			dot_legal = FALSE;
			sign_legal = FALSE;
		}
		if (*s == '-')
			sign_legal = FALSE;
		if (toupper(*s) == 'E') {
			exp_legal = FALSE;
			sign_legal = TRUE;
			dot_legal = FALSE;
			comma_legal = FALSE;
		}

		if (*s != ',') {
			s++;
			number_found = 1;
		}

		(*x1)++;
	}

	if (*x1 > x2) {
		*x1 = x1save;
		(*y1)++;
	}
	*s = 0;
	if (number_found)
		*value = atof(buf);
	else
		*value = 0.0;
	return(number_found);
}

void capture()
{
	int           x1save, x1, x2, y1, y2, i, op;
	double        tempf, xf;
	long          temp;

	x1 = old_col;
	y1 = old_row;
	op = get_region(&x1, &y1, &x2, &y2);

	if ((op == RIGHT_BUTTON) || (op == ESC))
		return;

	if (x2 < x1) { i = x1; x1 = x2; x2 = i; }
	if (y2 < y1) { i = y1; y1 = y2; y2 = i; }

	x1save = x1;

	if (get_value(x1save, &x1, &y1, x2, y2, &tempf))
		while (get_value(x1save, &x1, &y1, x2, y2, &xf))
			switch (op) {
				case MINUS:
					tempf -= xf;
					break;
				case MULT:
					tempf *= xf;
					break;
				case DIVIDE:
					if (xf != 0.0)
						tempf /= xf;
					break;
				case LEFT_BUTTON:
				case CR:
				case PLUS:
				default:
					tempf += xf;
					break;
			}
	temp = round(tempf);
	push(temp, tempf);
}

/**************************/
/* End of capture section */
/**************************/

#endif /* TSR */

/*****************************************/
/* Low level utility functions for DCALC */
/*****************************************/

void print_string(s)
char *s;
{
	int	i;

	wnwrbuf(cur_y, cur_x, 0, s, -1, -1, 1);
}

void gotoxy(x, y)
	int	x, y;
{
	wncurmov(cur_y=y, cur_x=x);
}

void msg(char *s)
{
	wndsplay(msgwin, &msglocation, &msgbord);
	wnselect(msgwin);
	gotoxy(ERRORX, ERRORY);
	strcpy(outbuf, s);
	while (strlen(outbuf) < MSG_SIZE - 1)
	/* don't write in last space because it makes the window scroll */
		strcat(outbuf, " ");
	outbuf[MSG_SIZE - 1] = '\0';
	print_string(outbuf);
	dispnums();
}

void clear_msg()
{
	wnremove(msgwin);
}

void put_a_char(COMMAND c)
{
	/* Only used to write to the X reg or ring the bell */
	wnselect(xwin);
	wnwrtty(c, -1, -1);
}

COMMAND get_a_char(COMMAND *d)
{
/* get a character with no echo, but with wait. If special, return SPECIAL
   with special value in d. Else d = 0.
*/
	union REGS inregs, outregs;
	unsigned char c;

	c = *d = 0;
	/* Function 0h for normal and 10h for extended keyboards */
	inregs.h.ah = peekb(0x40, 0x96) & 0x10;
	int86(0x16, &inregs, &outregs);
	/* Need to strip out extended keycodes if present */
	c = outregs.h.al;
	if (c == 0xe0)
		c = 0;
	*d = outregs.h.ah;
	if (c != 0)
		*d = 0;
	if (base != 1)
		c = toupper(c);
	return((COMMAND)c);
}

void print_fk(int i)
{
	int j;

	wnselect(fkwin);
	for (j = 0; j < 8; j++) {
		gotoxy(FK_X + j * 10, FK_Y);
		sprintf(outbuf, "%d:", j + 1);
		wnwrstr(outbuf, -1, -1);
		wnwrstr(func[i][j].label, 0, 7);
	}
}

COMMAND map_key_to_func(COMMAND c, COMMAND d)
{
	COMMAND ret_val;

	ret_val = c;

	/* ^@ is (0,3) */
	if ((c == SPECIAL) && (d != 3)) {
		if ((d >= F1KEY) && (d <= F8KEY))
			ret_val = func[fk_set][d - F1KEY].function;
		else switch (d) {
			case DOWNKEY:
				ret_val = ROLLDOWN;
				break;

			case UPKEY:
				ret_val = ENTER;
				break;

			case F10KEY:
			case NEXTKEY:
				ret_val = NEXTKEYS;
				break;

			case F9KEY:
			case PREVKEY:
				ret_val = PREVKEYS;
				break;

			case DELKEY:
				ret_val = CLX;
				break;

			default:
				break;
		}
	} else if (base != 1)
		if (c == ESCAPE)
			ret_val = QUIT;
		else if ((c == '\n') || (c == '\r'))
			ret_val = ENTER;

	return(ret_val);
}

pop_up_reg()
{
	int i;
	
	if (regwin && wngetopt(regwin, WN_DEVICE, &i))
		if (i < 0) {
			if (!wndsplay(regwin, &reglocation, &regbord))
				msg(reg_mem_msg);
		} else
			wnremove(regwin);
	 else
		 msg(reg_mem_msg);
	return;
}

void os_init()
{
	cursor_was_off = sccurst(&old_row, &old_col, &old_high, &old_low);
}

void os_term()
{
	wnremove(xwin);
	wnremove(regwin);
	wnremove(msgwin);
	wnremove(fkwin);
	wnremove(helpwin);
	sccurset(old_row, old_col);
	scpgcur(cursor_was_off, old_high, old_low, CUR_NO_ADJUST);
}

void w_fatal(void)
{
	extern int b_wnerr;

	fprintf(stderr, "Not enough memory\n");
	terminate();
}

void print_attr(char *s)
{
	while (*s) {
		if (*s == '\2')
			wnwrtty(*++s, 15, 0);
		else
			wnwrtty(*s, 7, 0);
		s++;
	}
}

void alloc_windows(void )
/* called from 'main' */
{
	if ((xwin =    wncreate( 2, WIDTH + 1, CYAN)) == NIL)
		w_fatal();
	if ((msgwin =  wncreate( 1, MSG_SIZE, CYAN)) == NIL)
		w_fatal();
	if ((fkwin =   wncreate( 1, SCREEN_MAX_X + 1, CYAN)) == NIL)
		w_fatal();

	regwin = wncreate(14, 38, CYAN); /* register window is not essential */

	if (helpwin = wncreate(10, 35, CYAN)) /* help is not essential */ {
		wnselect(helpwin);
		print_attr("\2E\2s\2c Exit DCALC     \2X x<>y\r\n");
		print_attr("\2@   Reg's On/Off   \2Y y^x\r\n");
		print_attr("\2M   Deg/Rad  \2P Pi  \2% x % of y\r\n");
		print_attr("\2D\2e\2l CLX            \2# y MODULO x\r\n");
		print_attr("\2\031   ROLL DN        \2& x AND y\r\n");
		print_attr("\2,   CHS            \2| x OR y \r\n");
		print_attr("\2L   Recall Last x  \2^ x XOR y\r\n");
		print_attr("\2S   Store          \2~ complement x\r\n");
		print_attr("\2R   Recall   \2>,\2< y shifted x bits\r\n");
		print_attr("\2F\002\071  Previous keys  \2F\002\061\002\060 Next keys");
	}
}

/* Pop up the three essential windows */
void pop_up_windows(void )
{
	int           mode, columns, act_page;

	xlocation.dev        = scmode(&mode, &columns, &act_page);
	xlocation.page       = act_page;
	xlocation.corner.row = 21;
	xlocation.corner.col = 2;
	if (wndsplay(xwin, &xlocation, &xbord) == NIL)
		w_fatal();

	msglocation.dev        = scmode(&mode, &columns, &act_page);
	msglocation.page       = act_page;
	msglocation.corner.row = 18;
	msglocation.corner.col = 1;
	if (wndsplay(msgwin, &msglocation, &msgbord) == NIL)
		w_fatal();

	fklocation.dev        = scmode(&mode, &columns, &act_page);
	fklocation.page       = act_page;
	fklocation.corner.row = SCREEN_MAX_Y;
	fklocation.corner.col = 0;
	if (wndsplay(fkwin, &fklocation, &fkbord) == NIL)
		w_fatal();

	reglocation.dev        = scmode(&mode, &columns, &act_page);
	reglocation.page       = act_page;
	reglocation.corner.row = 1;
	reglocation.corner.col = 1;

	helplocation.dev        = scmode(&mode, &columns, &act_page);
	helplocation.page       = act_page;
	helplocation.corner.row = 1;
	helplocation.corner.col = 42;
}

void pop_up_help()
{
	int i;

	if (helpwin && wngetopt(helpwin, WN_DEVICE, &i))
		if (i < 0) {
			if (!wndsplay(helpwin, &helplocation, &helpbord))
				msg(hlp_mem_msg);
		} else
			wnremove(helpwin);
	else
		msg(hlp_mem_msg);
}

xwin_select()
{
	return(wnselect(xwin) != NULL);
}

reg_select()
{
	return(wnselect(regwin) != NULL);
}

void matherr(x)
	struct exception *x;
{
	int	set_signals(),
		result;

	/* If xwin is there (ie not capturing) */
	if ((wngetopt(xwin, WN_DEVICE, &result) != NULL) &&
	    (result != ABSENT))
		msg("Math error!");
	set_signals();
	longjmp(save, 1);
	return; /* Never gets here - but, just in case! */
}

set_signals()
{
	signal(SIGFPE, matherr);
	signal(SIGINT, SIG_IGN);
}

void dcalc()
{
	COMMAND c, d;

	pop_up_windows();
	display();
	pop_up_help();
	pop_up_reg();
	msg(DEF_SIG);

	setjmp(save);
	set_signals();
	do {
		wnselect(xwin);
		wncursor(xwin);
		wncurmov(STARTX_Y, WIDTH);
		c = get_a_char(&d);
		clear_msg();
		c = map_key_to_func(c, d);
		process(c);
		setjmp(save);
	} while (c != QUIT);

	terminate();
	/* return to calling program - it will T&SR or exit */
}


void dispnums()
{
    xwin_select();
    gotoxy(LEFTSIDE, STARTX_Y);
    if (entering) {
	strcpy(outbuf, inbuf);
	prep_for_output(outbuf);
	print_string(outbuf);
    } else
	print_base(x, xf);
    
    if (reg_select()) {
	gotoxy(LEFTSIDE, STARTT_Y);
	print_string("T");
	gotoxy(START_X, STARTT_Y);
	print_base(t, tf);
	
	gotoxy(LEFTSIDE, STARTZ_Y);
	print_string("Z");
	gotoxy(START_X, STARTZ_Y);
	print_base(z, zf);
	
	gotoxy(LEFTSIDE, STARTY_Y);
	print_string("Y");
	gotoxy(START_X, STARTY_Y);
	print_base(y, yf);
	
	gotoxy(LEFTSIDE, STARTL_Y);
	print_string("L");
	gotoxy(START_X, STARTL_Y);
	print_base(l, lf);
    }
}
  
void dispreg(i)
    int	i;
{
    if (reg_select()) {
	gotoxy(LEFTSIDE, i);
	outbuf[0] = '0' + i;
	outbuf[1] = '\0';
	print_string(outbuf);
	gotoxy(START_X, i);
	print_base(reg[i], regf[i]);
    }
}

void dispregs()
{
    int i;
    
    for (i = 0; i < NUMREGS; i++)
	dispreg(i);
}

void prinbase()
{
    xwin_select();
    gotoxy(BASE_X, BASE_Y);
    switch (base) {
    case 1:
	print_string("Ascii   ");
	break;
	
    case 2:
	print_string("Binary  ");
	break;
	
    case 8:
	print_string("Octal   ");
	break;
	
    case 10:
	print_string("Decimal ");
	break;
	
    case 16:
	print_string("Hex     ");
	break;
	
    default:
	if (base < -9) {
	    sprintf(outbuf, "Engin. %d", -(base + 10));
	    print_string(outbuf);
	} else {
	    sprintf(outbuf, "Float  %d", -base);
	    print_string(outbuf);
	}
	break;
    }
}

void print_inv()
{
    xwin_select();
    gotoxy(INV_X, INV_Y);
    if (invert)
	print_string("Inv   ");
    else
	print_string("      ");
}

void print_base(x, xf)
    long	x;
    double	xf;
{
    fmt_base(outbuf, x, xf);
    prep_for_output(outbuf);
    print_string(outbuf);
}

void print_deg()
{
    xwin_select();
    gotoxy(TMODE_X, TMODE_Y);
    if (degree)
	print_string("Degrees");
    else
	print_string("Radians");
}

void print_x(char *buf) 
{
    gotoxy(LEFTSIDE, STARTX_Y);
    print_string(buf);
}

void main(argc, argv)
	int   argc;
	char *argv[];
{
	int i;

	alloc_windows(); /* Pre-alloc windows in case we go TSR */
	initialise();	 /* Initialise calculator */

#ifdef TSR
	/* Process arguments */
	argc--; *argv++;
	while (argc) {
		if ((**argv == '-') || (**argv == '/')) {
			(*argv)++;
			**argv = toupper(**argv);
			switch(**argv) {
				case 'R':
					exit(remove_iv());
					break;

				case 'I':
					exit (install_iv ());
					break;
				default:
					exit(usage());
					break;
			}
		}
		argc--; argv++;
	}
#endif TSR
	dcalc();
	exit(0);
}

#endif /* __TURBOC__ */

#endif /* PCSPECIFIC */

